"""Module containing the computation of p_astro by source category
"""
import itertools
from copy import deepcopy

import numpy as np
from scipy.special import la_roots
from numpy.polynomial.hermite import hermgauss


class CountPosteriorElements(object):
    '''
    Class that collects information of the candidate events on
    which the FGMC counts posterior is to be built. This information
    is not specific to a source type.
    '''

    def __init__(self, f_divby_b, buff=1e-15, verbose=True):
        '''
        Parameters
        ----------
        f_divby_b : list, tuple
            iterable of real numbers, foreground to background ratio
        buff : float
            Surrogate for zero, to avoid singularities in the
            counts posterior. Default value 1e-15
        verbose : bool
        '''
        self.f_divby_b = f_divby_b
        self.buffer = buff
        self.verbose = verbose


class SourceType(object):
    '''
    Class that collects source-type specific information of
    candidate events. Instances of this class will be passed
    to the :class:`Posterior` class to instantiate the latter.
    '''

    def __init__(self, label, w_fgmc):
        '''
        Parameters
        ----------

        label : str
            label associated with a class instance, like, "BNS", "bns", "BBH"
        w_fgmc : float
            In the **course-grained method**, `w_fgmc` is defined as
            p(SVD bin #|source-type), derived from the distribution of
            recovered parameters across SVD bins during an injection
            campaign targeted at a source-type. In the **fine-grained method**,
            `w_fgmc` is defined as p(template|source-type).
        '''
        self.label = label
        self.w_fgmc = w_fgmc


class Posterior(CountPosteriorElements):
    '''
    Class that constructs the **FGMC** counts posterior.

    Attributes
    ----------
    args_fixed : list
        list of object instances whose labels match
        keys of ``fix_sources`` dictionary.
    args_free : list
        list of object instances whose labels do not
        match the keys of fix_sources dictionary.
    arg_terr : :class:`SourceType`
        pertaining to the terrestrial source
    args_astro : tuple
        instances of :class:`SourceType` that correspond to
        astrophysical categories.
    keys_all : list
        list of labels of instances of all :class:`SourceType` instances.
    key_terr : list
        subset of ``key_all`` with terrestrial source type
    keys_astro : list
        subset of ``key_all`` with astrophysical source type
    keys_fixed : list
        list of labels of pinned :class:`SourceType` instances.
    keys_free : list
        list of labels of free :class:`SourceType` instances.
    n : int
        number of terms in the `gauss-laguerre` and `gauss-hermite`
        quadrature integrals, hard coded to 20.
    N : int
        number of candidate events
    roots : iter
        abcissa of `gauss-laguerre` quadrature
    weights : iter
        weights of `gauss-laguerre` quadrature
    hroots : iter
        abcissa of `gauss-hermite` quadrature
    hweights : iter
        weights of `gauss-hermite` quadrature
    reg_const : float
        regularization constant to avoid numerical over/under flows
    norm : float
        normalization constant for counts posterior, determined
        post regularization
    terr_jacobian : float
        constant to be tacked on when integrating terrestrial counts via
        `gauss-hermite` quadrature
        This can only be done after variable transformation.
    priorDict : dict
        dictionary of priors keyed on *Uniform* and *Jeffreys*
    '''
    def __init__(self, f_divby_b, prior_type, terr_source,
                 fix_sources={},
                 buff=1e-15,
                 mcmc=False,
                 verbose=True,
                 **astro_sources):
        '''
        Parameters
        ----------
        f_divby_b : list, tuple
            iterable of real numbers, foreground to background ratio
        prior_type : str
            **Uniform** or **Jeffreys**
        terr_source : :class:`SourceType`
            pertaining to the terrestrial source
        fix_sources : dict
            fixed lambda values, keyed on source labels.

            Example: :code:`{"Terr":value}`
            Note that the keys *must* match the labels
            used for :class:`SourceType` instances.
        buff : float
            Surrogate for zero, to avoid singularities in the
            counts posterior.
        verbose : bool
        astro_sources : dict
            dictionary of :class`:SourceType, pertaining to
            astrophysical sources
        '''

        CountPosteriorElements.__init__(self, f_divby_b, buff, verbose)

        assert (astro_sources != {}), \
            "Need at least one astrophysical source."
        assert (terr_source != {}), "Terrestrial source required."

        self.args_astro = astro_sources.values()
        self.arg_terr = terr_source
        self.args_all = list(self.args_astro) + [self.arg_terr]
        self.keys_all = [arg.label for arg in self.args_all]
        self.key_terr = self.arg_terr.label
        self.keys_astro = [arg.label for arg in self.args_astro]
        self.fix_sources = fix_sources
        self.keys_fixed = self.fix_sources.keys()
        self.mcmc = mcmc

        assert (
            len(self.f_divby_b) ==
            int(np.sum([len(arg.w_fgmc) for arg in self.args_all]) /
                len(self.args_all))), \
            "Lengths of f_divby_b vector, and w_fgmc vectors, must match"
        assert (self.key_terr not in self.keys_astro), \
            ("Terrestrial SourceType must be input separately, "
             "via terr_source field")
        if self.fix_sources != {}:
            assert (
                len(set(self.keys_all) & set(self.keys_fixed)) ==
                len(self.keys_fixed)), \
                ("Label(s) of fix_sources don't match labels of "
                 "astro/terr source instances")

        self.args_fixed = \
            [list(filter(lambda x: x.label == label, self.args_all))[0] for
             label in self.keys_fixed]

        self.args_free = list(set(self.args_all)-set(self.args_fixed))
        self.keys_free = [arg.label for arg in self.args_free]
        self.prior_type = prior_type
        self.n = 20
        self.N = len(list(self.args_astro)[0].w_fgmc)
        self.terr_jacobian = np.sqrt(2*self.N)

        if self.prior_type == "Uniform":
            self.roots, self.weights = la_roots(n=self.n, alpha=0)
        if self.prior_type == "Jeffreys":
            self.roots, self.weights = la_roots(n=self.n, alpha=-0.5)
        self.hroots, self.hweights = hermgauss(deg=self.n)

        if not self.mcmc:
            self.norm_array = self.regularize()
            self.reg_const = np.max(self.norm_array)
            self.norm = self.normalize()

        self.priorDict = {}
        self.priorDict["Uniform"] = Posterior.priorUniform
        self.priorDict["Jeffreys"] = Posterior.priorJeffreys

        if self.verbose:
            print("Posterior class instantiation complete.")

    def reduced_log_likelihood_terr(self, **lambdas):
        '''
        The log of the FGMC posterior divided by
        (np.exp(-np.sum(lambdas))*lambda_terr**N)
        Takes the lambdas as input variables. One of the lambdas
        *must* be ``lambda_terr``.
        Also, the keys on lambdas *must* match the labels of the
        :class:`SourceType` instances.

        Evaluating the reduced posterior at ``lambda_bns = 5``
        will need to be passed as BNS=5, if "BNS" is the label of
        the corresponding instance of SourceType.
        Returns the value of the reduced log posterior for a
        set of lambda values.
        '''
        return np.sum(np.log(self.arg_terr.w_fgmc +
                      np.outer(1/lambdas[self.key_terr], self.f_divby_b) *
                      np.sum([np.outer(lambdas[arg.label], arg.w_fgmc)
                             for arg in self.args_astro], axis=0)), axis=1)

    def reduced_log_likelihood_astro(self, **lambdas):
        '''
        The log of the FGMC posterior divided by
        (np.exp(-np.sum(lambdas))*((N**N)*exp(-N)))
        Takes the lambdas as input variables. One of the
        lambdas *must* be ``lambda_terr``.
        Also, the keys on lambdas *must* match the labels of
        the :class:`SourceType` instances.
        So, evaluating the reduced posterior at ``lambda_bns = 5``
        will need to be passed as BNS=5, if "BNS" is the label of
        the corresponding instance of SourceType. Returns
        the value of the reduced log posterior for a set of lambda values.
        '''
        return (self.reduced_log_likelihood_terr(**lambdas) +
                self.N*np.log(lambdas[self.key_terr]) -
                self.N*(np.log(self.N) - 1))

    def TerrVarTransform(self, x, shift_amount="Default"):  # noqa: N802
        '''
        Variable transformation for lambda_terr. Used when
        marginalizing over ``lambda_terr`` via Gauss-Hermite quadrature.

        Parameters
        ----------
        x : float
            Variable to be transformed
        shift_amount : float
            If shift amount is ``a``, function will transform
            a Gaussian with ``mu=a``, ``sigma=np.sqrt(a)``
            to a Gaussian that goes like ``exp(-x**2)``

        Returns
        -------
        float
            transformed variable
        '''
        if shift_amount == "Default":
            return x*np.sqrt(2*self.N)+self.N
        else:
            return x*np.sqrt(2*shift_amount)+shift_amount

    def marginalize_gq(self, func, categories,
                       shift_amount="Default", **pinned_lambdas):
        '''
        Function to marginalize posterior on counts using Gaussian quadrature
        Gauss-Laguerre for marginalization over astrophysical counts
        Gauss-Hermite for marginalization over terrestrial counts

        Notes
        -----
            Astrophyical Counts Marginalization (Gauss-Laguerre Quadrature):

                :math:`f(x) = P(x)e^{-x}`
                
                :math:`\int_{0}^{\infty}f(x)dx = \sum_{i=1}^{n}w_iP(r_i)
                = \sum_{i=1}^{n}e^{\log(w_i)+\log(P(r_i))}`

            where :math:`w_i, r_i` are the Gauss-Laguerre weights and abscissa.

            Terrestrial Counts Marginalization (Gauss-Hermite):

                :math:`f(x) = P(x)e^{-x^2}`
            
                :math:`\int_{-\infty}^{+\infty}f(x)dx = \sum_{i=1}^{n}w_iP(r_i)
                = \sum_{i=1}^{n}e^{\log(w_i)+\log(P(r_i))}`

            where :math:`w_i, r_i` are the Gauss-Hermite weights and abscissa.

        Parameters
        ----------

        func : callable
            log of function to be marginalized divided by the Gauss-Laguerre
            weighting factor.
        categories : list
            list of source types to be marginalized over. Strings *must*
            match labels of instances of SourceType.
        shift_amount : float
            same as shift amount in :meth:`TerrVarTransform`,
            used for marginalizing over lambda_terr
        pinned_lambdas : dict
            values of lambdas that are not to be marginalized over.
            keys *must* match labels of instances of SourceType class.

        Returns
        -------

        list
            list of values, which when exponentiated and summed,
            give the value of the marginalized posterior.
        '''

        thelist = []  # Initialize list
        lambdas = deepcopy(pinned_lambdas)  # Set dictionary to pinned_lambdas

        # Iterate over range of quadrature roots & weights terms, for every
        # category to be marginalized over
        for idx in itertools.product(range(self.n), repeat=len(categories)):
            sum_log_weights = 0  # initialize log of quadrature weights
            for key, root, weight, hroot, hweight in zip(
                    categories,
                    self.roots[list(idx)],
                    self.weights[list(idx)],
                    self.hroots[list(idx)],
                    self.hweights[list(idx)]):
                # Use Gauss-Hermite quadrature to integrate terrestrial counts
                if key == self.key_terr:
                    lambdas[key] = self.TerrVarTransform(hroot, shift_amount)
                    sum_log_weights += np.log(hweight)
                # Gauss-Laguerre quadrature to integrate astrophysical counts
                else:
                    lambdas[key] = root
                    sum_log_weights += np.log(weight)
            # exponentiate and sum to give the marginalization value
            thelist.append(func(**lambdas)+sum_log_weights)

        if self.key_terr in categories:
            return np.array(thelist)+np.log(self.terr_jacobian)
        else:
            return np.array(thelist)

    def regularize(self):
        '''
        Function to regularize FGMC posterior to avoid numerical
        over-/under-flows.
        It simply invokes marginalize_gq and applies it on the reduced
        log posterior. The regularization constant is simply the max value
        of the array returned by this function.
        '''
        if self.verbose:
            print("Regularizing FGMC posterior to avoid ",
                  "numerical over-/under-flows ...")
        if self.key_terr in self.keys_free:
            return self.marginalize_gq(
                self.reduced_log_likelihood_terr,
                self.keys_free,
                **self.fix_sources)
        else:
            return self.marginalize_gq(
                self.reduced_log_likelihood_astro,
                self.keys_free,
                **self.fix_sources)

    def normalize(self):
        '''
        Function to normalize FGMC posterior.
        This is simply the sum of the array (exponentiated)
        returned by the regularize function,
        with the regularization constant removed.
        '''
        if self.verbose:
            print("Normalizing FGMC posterior ...")
        return np.sum(np.exp(self.norm_array-self.reg_const))

    def poissonLogWeight(self, **lambdas):  # noqa: N802
        '''
        Function that returns the log of the Poisson weights: log(e^(-x)) = -x

        Parameters
        ----------
        lambdas : dict
            dictionary of counts and their values

        Returns
        -------
        float
            log of the Poisson weights
        '''
        return -np.sum(np.array(list(lambdas.values())), axis=0)

    def gaussianLogWeight(self, terr_count):  # noqa: N802
        '''
        Function that returns the log of the Hermite weights:
        log(e^(-(x-N)^2/(2N)))

        Parameters
        ----------

        terr_count : float
            terrestrial count value

        Returns
        -------

        float
            log of the Gaussian with mean and variance N
        '''

        return -(terr_count-self.N)**2/(2*self.N)

    @staticmethod
    def priorUniform(**lambdas):  # noqa: N802
        '''
        Uniform Prior on Counts. Takes dictionary of
        counts. Keys *must* match labels of instances
        of SourceType class.
        '''
        return 1

    @staticmethod
    def priorJeffreys(**lambdas):  # noqa: N802
        '''
        Jeffreys Prior on Counts. Takes dictionary of
        counts. Keys *must* match labels of instances
        of SourceType class.
        '''
        return 1/(np.prod(np.sqrt(lambdas.values()), axis=0))

class MarginalizedPosterior(Posterior):

    '''
    Class that provides the marginalized FGMC posterior on counts, 
    for any number of dimensions up to the max dimensions allowed
    by the original multidimensional FGMC posterior. Inherits from
    the Posterior class.
    Computes astrophysical probability p_astro, by source category.
    '''

    '''
    Attributes
    ----------
    
    args_fixed : list
        list of object instances whose labels match
        keys of ``fix_sources`` dictionary.
    args_free : list
        list of object instances whose labels do not
        match the keys of fix_sources dictionary.
    arg_terr : :class:`SourceType`
        pertaining to the terrestrial source
    args_astro : tuple
        instances of :class:`SourceType` that correspond to
        astrophysical categories.
    keys_all : list
        list of labels of instances of all :class:`SourceType` instances.
    key_terr : list
        subset of ``key_all`` with terrestrial source type
    keys_astro : list
        subset of ``key_all`` with astrophysical source type
    keys_fixed : list
        list of labels of pinned :class:`SourceType` instances.
    keys_free : list
        list of labels of free :class:`SourceType` instances.
    n : int
        number of terms in the `gauss-laguerre` and `gauss-hermite`
        quadrature integrals, hard coded to 20.
    N : int
        number of candidate events
    roots : iter
        abcissa of `gauss-laguerre` quadrature
    weights : iter
        weights of `gauss-laguerre` quadrature
    hroots : iter
        abcissa of `gauss-hermite` quadrature
    hweights : iter
        weights of `gauss-hermite` quadrature
    reg_const : float
        regularization constant to avoid numerical over/under flows
    norm : float
        normalization constant for counts posterior, determined
        post regularization
    terr_jacobian : float
        constant to be tacked on when integrating terrestrial counts via
        `gauss-hermite` quadrature
        This can only be done after variable transformation.
    priorDict : dict
        dictionary of priors keyed on *Uniform* and *Jeffreys*
    '''

    def __init__(self,f_divby_b,prior_type,terr_source,fix_sources={},
                 buff=1e-15,mcmc=False,verbose=True,**astro_sources):
        
        '''
        Parameters
        ----------
        
        f_divby_b : list, tuple
            iterable of real numbers, foreground to background ratio
        prior_type : str
            **Uniform** or **Jeffreys**
        terr_source : :class:`SourceType`
            pertaining to the terrestrial source
        fix_sources : dict
            fixed lambda values, keyed on source labels.
            Example: :code:`{"Terr":value}`
            Note that the keys *must* match the labels
            used for :class:`SourceType` instances.
        buff : float
            Surrogate for zero, to avoid singularities in the
            counts posterior.
        verbose : bool
        astro_sources : dict
            dictionary of :class`:SourceType`, pertaining to
            astrophysical sources
        '''

        Posterior.__init__(self,f_divby_b,prior_type,terr_source,fix_sources,buff,mcmc,verbose,**astro_sources)
        if self.verbose:
            print ("MarginalizedPosterior class instantiation complete.")
    
    def posterior(self,**lambdas):

        '''
        FGMC Counts Posterior function. Takes dictionary of
        counts. Keys *must* match labels of instances
        of :class:`SourceType`.
        
        Posterior can have any dimensionality up to max allowable
        dimensionality, as determined by the number of SourceType 
        instances passed to the MarginalizedPosterior class.
        
        Parameters
        ----------
        
        lambdas : dict
            counts:value dictionary pairs.
            Keys *must* match labels of instances of SourceType
            class.
        
        Returns
        -------

        float
            posterior evaluated at values supplied in the lambda dictionary
        
        Example: :code:`posterior(BNS=5)` 
        This will give the value for the 
        corresponding 1-dimensional marginalized posterior

        Example: :code:`posterior(BBH=1,NSBH=3)`
        This will give the value for the 
        corresponding 2-dimensional marginalized posterior.
        
        '''
        
        assert(len(set(self.keys_all)&set(lambdas.keys()))==len(lambdas.keys())),"Counts label(s) passed to posterior function don't match labels of astro/terr source instances"
        
        # keys corresponding to sources to be marginalized over
        self.keys_marg = set(self.keys_free)-set(lambdas.keys())
        # counts corresponding to sources that are pinned
        pinned_lambdas = deepcopy(self.fix_sources)
        pinned_lambdas.update(lambdas)
        # initialize list
        thelist = []

        # If there are no sources to marginalize over
        if len(self.keys_marg) == 0:
            log_likelihood = self.poissonLogWeight(**lambdas)+self.reduced_log_likelihood_astro(**pinned_lambdas)
            likelihood = np.exp(log_likelihood-self.reg_const)
            return self.priorDict[self.prior_type](**lambdas)*likelihood/self.norm
        else:
            # If one of the sources to marginalize over is terrestrial
            if self.key_terr in self.keys_marg:
                thelist = self.marginalize_gq(self.reduced_log_likelihood_terr,self.keys_marg,**pinned_lambdas)
            # If none of the sources to marginalize over is terrestrial
            else:
                thelist = self.marginalize_gq(self.reduced_log_likelihood_astro,self.keys_marg,**pinned_lambdas)
            # Regularize and normalize
            thelist += self.poissonLogWeight(**lambdas)-np.log(self.norm)-self.reg_const
            return self.priorDict[self.prior_type](**lambdas)*np.sum(np.exp(thelist),axis=0)
        
    def mean(self,categories,posterior="Default",fix_sources="Default"):

        '''
        Function to determine mean of posterior.
        
        Parameters
        ----------
        
        categories: list of strings
            Keys *must* match labels of instances of :class:`SourceType`
        posterior: callabel function 
            User-defined counts posterior function. Must
            be constructed in the same way as the posterior function defined in this
            class, i.e, must take as input a dictionary of counts keyed on labels, etc.
            The "Default" is to use the posterior function defined in this class. 
        fix_sources: dictionary 
            dictionary of pinned sources, with source type
            labels as keys and numbers (to pin to) as values. Defaults to fix_sources
            supplied to instantiate this class.
        
        
        Returns
        -------
        
        float: 
            mean of posterior
        
        Example :code:`mean(["BNS"])` 
        returns <lambda_BNS>
        Example :code:`mean(["BNS","BBH"])` 
        returns <lambda_BNS*lambda_BBH>
        etc.
        '''
        
        # Dictionary, indexed on source category, 
        # that counts the number of times a source type 
        # has been repeated
        cat_count_dict = {x:categories.count(x) for x in categories}
        
        assert(len(set(self.keys_all)&set(categories))==len(cat_count_dict.keys())),"Counts label(s) passed to mean function don't match labels of the source instances"
        
        if posterior == "Default":
            posterior = self.posterior
            
        if fix_sources == "Default":
            fix_sources = self.fix_sources
        
        # Extract pinned sources from categories passed to mean function
        fixed_cat = list(set(categories)&set(self.keys_fixed))
        
        # Determine product of values of fixed_cat sources
        if fixed_cat == []:
            pinned_mean = 1
        else:
            pinned_mean = np.prod([fix_sources[cat]**cat_count_dict[cat] for cat in fixed_cat])
        
        # If all categories are pinned sources, return product of values
        if len(fixed_cat) == len(categories):
            return pinned_mean
        
        # Remove fixed categories from categories passed to mean function
        free_cat = list(set(categories)-set(fixed_cat))
        
        # Determine if terrestrial counts is to be integrated over
        terr_exists = self.key_terr in free_cat

        if terr_exists: 
            func = lambda **x: (np.sum(np.log([x[cat]**cat_count_dict[cat] for cat in free_cat]))+
                                np.log(posterior(**x))-
                                np.log(self.priorDict[self.prior_type](**x))-
                                self.poissonLogWeight(**x)-
                                (self.N*np.log(x[self.key_terr])-
                                self.N*(np.log(self.N)-1)))
        else:
            func = lambda **x: (np.sum(np.log([x[cat]**cat_count_dict[cat] for cat in free_cat]))+
                                np.log(posterior(**x))-
                                np.log(self.priorDict[self.prior_type](**x))-
                                self.poissonLogWeight(**x))
        
        return pinned_mean*np.sum(np.exp(self.marginalize_gq(func,free_cat)))
    
    def getOneDimMean(self,category,posterior="Default",fix_sources="Default"):

        '''
        Function to determine mean of 1-dimensional marginalized posterior on counts
        
        Parameters
        ----------
        
        category: string 
            Source category whose marginalized counts posterior
            is sought. *Must* match one of the labels of non-pinned sources. 
        posterior: callable function 
            User-defined counts posterior function. Must
            be constructed in the same way as the posterior function defined in this
            class, i.e, must take as input a dictionary of counts keyed on labels, etc.
            The "Default" is to use the posterior function defined in this class. 
        fix_sources: dictionary 
            dictionary of pinned sources, with source type
            labels as keys and numbers (to pin to) as values. Defaults to fix_sources
            supplied to instantiate this class.
        
        Returns
        -------
        
        float:
            mean of 1D marginalized posterior
        
        '''

        assert(category in self.keys_all),"Category passed to getOneDimMean function doesn't match any labels of the source instances"
        
        if posterior == "Default":
            posterior = self.posterior
            
        if fix_sources == "Default":
            fix_sources = self.fix_sources
        
        return self.mean([category],posterior,fix_sources)
    
    
    def getCovariance(self,categories,posterior="Default",fix_sources="Default"):

        '''
        Function to determine covariance of counts
        
        Parameters
        ----------
        
        categories: list of strings
            Keys must match labels of instances of :class:`SourceType`
        posterior: callable function 
            User-defined counts posterior function. Must
            be constructed in the same way as the posterior function defined in this
            class, i.e, must take as input a dictionary of counts keyed on labels, etc.
            The "Default" is to use the posterior function defined in this class. 
        fix_sources: dictionary
            Dictionary of pinned sources, with source type
            labels as keys and numbers (to pin to) as values. Defaults to fix_sources
            supplied to instantiate this class.
        
        Returns
        -------
        
        float:
            covariance between Poisson expected counts
        '''
        
        assert(len(categories) == 2),"Covariance functionality available for only two categories"
        
        if posterior == "Default":
            posterior = self.posterior
            
        if fix_sources == "Default":
            fix_sources = self.fix_sources
            
        return (self.mean(categories,posterior,fix_sources)-
                        np.prod([self.mean([cat],posterior,fix_sources) for cat in categories]))
    

    def pastro_numerator(self,trigger_idx,categories,**lambdas):
        '''
        Function returns :math:`\sum_{\\alpha}\Lambda_{\\alpha}w_{\\alpha}(j)\\frac{f(j)}{b(j)}`
        
        for a trigger j.
        
        Parameters
        ----------
        
        trigger_idx: int or array of ints: 
            index (indices) of trigger(s) whose p(astro) value is sought
            indices map to the foreground/background values in the f_divby_b vector supplied
        categories:list
            list of category labels. Labels *must* match
            labels of instances of SourceType class
        lambdas:dict
            Counts dictionary. Keys *must* match
            labels of instances of :class:`SourceType`.
        
        Returns
        -------
        
        float:
            Value of pastro_numerator evaluated at lambdas.values()
        
        '''
        
        # Distill instances corresponding to categories list
        args = [list(filter(lambda x: x.label == label, self.args_all))[0] for label in categories]
        # Return \Sigma_{\alpha}lambda_{\alpha}*w_{\alpha}(j)*f(j)/b(j)
        # for \alpha iterated over categories provided
        return (np.sum([lambdas[arg.label]*arg.w_fgmc[trigger_idx]
                       for arg in args],axis=0)*self.f_divby_b[trigger_idx])
    
    def pterr_numerator(self,trigger_idx,lambda_terr):
        '''
        Function returns :math:`\Lambda_{\mathrm{terr}}w_{\mathrm{terr}}(j)`
        for a trigger j.
        
        Parameters
        ----------
        
        trigger_idx: int or array of ints: 
            index (indices) of trigger(s) whose p(astro) value is sought
            indices map to the foreground/background values in the f_divby_b vector supplied
        lambda_terr:float
            value of terrestrial count
        
        Returns
        -------
       
        float:
            Value of pterr_numerator evaluated at lambda_terr 
        
        '''
        return lambda_terr*self.arg_terr.w_fgmc[trigger_idx]
    
    def p_numerator(self,trigger_idx,categories,**lambdas):
 
        '''
        Function combines pastro_numerator and pterr_numerator,
        depending on list of categories over which p_astro is
        to be computed

        Parameters
        ----------

        trigger_idx: int or numpy array of ints: 
            index (indices) of trigger(s) whose p(astro) value is sought
            indices map to the foreground/background values in the f_divby_b vector supplied
        categories:list
            list of category labels. Labels *must* match
            labels of instances of SourceType class
        lambdas:dict
            Counts dictionary. Keys *must* match
            labels of instances of SourceType class. 
        
        Returns
        -------
        
        float or numpy array of floats:
            Value of p_numerator evaluated at lambdas.values()
        '''
        
        val = 0 # initialize p_numerator value
        if self.key_terr in categories:
            # Add pterr_numerator if terrestrial category is supplied
            val += self.pterr_numerator(trigger_idx,lambdas[self.key_terr])
            # Return value if *only* terrestrial category is supplied
            if len(lambdas.keys()) == 1:
                return val
        categories_astro = list(set(categories)-set([self.key_terr]))
        # Add pastro_numerator for all astrophysical categories supplied
        val += self.pastro_numerator(trigger_idx,categories_astro,**lambdas)
        return val       
    
    def p_denominator(self,trigger_idx,**lambdas):

        '''
        Function returns :math:`\Lambda_{\mathrm{terr}}w_\mathrm{terr}(j)+\\frac{f(j)}{b(j)}\sum_{\\alpha}\Lambda_{\\alpha}w_{\\alpha}(j))`
        
        Parameters
        ----------
        
        trigger_idx: int or numpy array of int
            index of trigger whose p_astro value is sought
        lambdas: dict 
            Counts dictionary. Keys *must* match
            labels of instances of SourceType class.
        
        Returns
        -------
        
        int or numpy array of ints
            value of p_denominator  
        '''

        return lambdas[self.key_terr]*(self.arg_terr.w_fgmc[trigger_idx]+
                       self.f_divby_b[trigger_idx]/lambdas[self.key_terr]*np.sum([lambdas[arg.label]*
                       arg.w_fgmc[trigger_idx] for arg in self.args_astro],axis=0))
    
    def p_integrand(self,trigger_idx,categories,**lambdas):

        '''
        Function combines `meth`:p_numerator and `meth`:p_denominator,
        along with posterior, to construct the integrand for p(astro).

        Parameters
        ----------
       
        trigger_idx: int or numpy array of ints: 
            index (indices) of trigger(s) whose p(astro) value is sought
            indices correspond to those of f_divby_b numpy array
        categories:list
            list of category labels. Labels *must* match
            labels of instances of `class`:SourceType.
        lambdas:dict
            Counts dictionary. Keys *must* match
            labels of instances of `class`:SourceType.  
        
        Returns
        -------
        
        float or numpy array of floats:
            Value of p_integrand       
        '''
        
        pinned_lambdas = deepcopy(self.fix_sources)
        if self.key_terr not in pinned_lambdas.keys():
            return (self.reduced_log_likelihood_terr(**lambdas)+
                   np.log(self.p_numerator(trigger_idx,categories,**lambdas))-
                   np.log(self.p_denominator(trigger_idx,**lambdas)))
        else:
            return (self.reduced_log_likelihood_astro(**lambdas)+
                   np.log(self.p_numerator(trigger_idx,categories,**lambdas))-
                   np.log(self.p_denominator(trigger_idx,**lambdas)))
    
    def pastro(self,trigger_idx,categories):

        '''
        Function invokes marginalize_gq to integrate p_integrand
        and thus determine p(astro)
        
        Parameters
        ----------
        
        trigger_idx: int or numpy array of ints
            index (indices) of trigger(s) whose p_astro value is sought
            indices correspond to those of f_divby_b numpy array
        categories: list 
            Category labels. These labels *must* match
            those of the instances of the :class:`SourceType`.
        
        Returns
        -------
        
        float or numpy array of floats:
            Value of p_astro
        
        '''
        
        assert(len(set(self.keys_all)&set(categories)) == len(categories)),"categories passed to pastro function don't match labels of the astro/terr source instances"
        
        pinned_lambdas = deepcopy(self.fix_sources)
        func = lambda **x: self.p_integrand(trigger_idx,categories,**x)
        thelist = self.marginalize_gq(func,self.keys_free,**pinned_lambdas)
        thelist += -np.log(self.norm) -self.reg_const
        return np.sum(np.exp(thelist),axis=0)

    
    def pastro_update(self,categories,bayesfac_dict,mean_values_dict):

        '''
           Function returns pastro of new candidate event(s),
           given posterior constructed from events not involving
           the new candidate event(s).
        
        Parameters
        ----------
        
        categories: list
            Category labels. These labels *must* match
            those of the instances of the SourceType
            class.       
        bayesfac_dict: dictionary
            Bayesfactor values of the new candidate event,
            keyed on SourceType labels. There should be a
            value associated with each astrophysical SourceType
            used to construct the FGMC posterior.
        mean_values_dict: dictionary
            Mean values of Poisson expected counts of each
            astrophysical SourceType. There should be a
            value associated with each astrophysical SourceType
            used to construct the FGMC posterior.
            
        
        Returns
        -------
        
        float or numpy array of floats
            Value of p_astro
            
        '''
        
        assert(len(set(self.keys_all)&set(categories)) == len(categories)),"categories passed to pastro function don't match labels of the astro/terr source instances"
        assert(set(self.keys_all) == set(mean_values_dict.keys())), "keys of mean_values_dict don't match labels of the astro/terr source instances"
        assert(set(self.keys_astro) == set(bayesfac_dict.keys())), "keys of bayesfac_dict don't match labels of the astro source instances"
        
        if self.key_terr in categories:
            num_terr = mean_values_dict[self.key_terr]
        else:
            num_terr = 0
        
        categories_astro = list(set(categories)-set([self.key_terr]))
        numerator = num_terr + np.sum([mean_values_dict[key]*bayesfac_dict[key] for key in categories_astro], axis=0)
        denominator = mean_values_dict[self.key_terr] + np.sum([mean_values_dict[key]*bayesfac_dict[key] 
                                                           for key in self.keys_astro], axis=0)
        
        return numerator/denominator



import io
import json

from ligo.lw import ligolw
from ligo.lw.ligolw import LIGOLWContentHandler
from ligo.lw import array as ligolw_array
from ligo.lw import param as ligolw_param
from ligo.lw import utils as ligolw_utils
from ligo.lw import lsctables
from ligo.segments import utils as segments_utils
from lal import rate

import itertools
import math
import numpy as np
from scipy import interpolate
from scipy import optimize

def assert_probability(f):
    """
    Adapted from gstlal/stats, gstlal-inspiral-1.6 branch
    """
    def g(*args, **kwargs):
        p = f(*args, **kwargs)
        if isinstance(p, np.ndarray):
            assert ((0. <= p) & (p <= 1.)).all()
        else:
            assert 0. <= p <= 1.
        return p
    return g

@assert_probability
@np.vectorize
def poisson_p_not_0(var):
    """
    Adapted from gstlal/stats, gstlal-inspiral-1.6 branch
    """
    assert var >= 0.
    var = -float(var)

    if var < -0.69314718055994529:
        return 1. - math.exp(var)

    s = [var]
    term = var
    threshold = -1e-20 * var
    for n in itertools.count(2):
        term *= var / n
        s.append(term)
        if abs(term) <= threshold:
            s.reverse()
            s = -sum(s)
            assert s >= 0.
            return s if s else 0.


class _RankingStatPDF(object):
    """
    Adapted from far.py, gstlal-inspiral-1.6 branch
    """
    ligo_lw_name_suffix = "gstlal_inspiral_rankingstatpdf"

    @classmethod
    def from_xml(cls, xml, name):
        """
        Find the root of the XML tree containing the serialization of this
        object
        """
        xml, = [elem for elem in
                xml.getElementsByTagName(ligolw.LIGO_LW.tagName)
                if elem.hasAttribute("Name") and
                elem.Name == "%s:%s" % (name, cls.ligo_lw_name_suffix)]
        # create a uninitialized instance
        self = super().__new__(cls)
        # populate from XML
        self.noise_lr_lnpdf = rate.BinnedLnPDF.from_xml(xml, "noise_lr_lnpdf")
        self.signal_lr_lnpdf = rate.BinnedLnPDF.from_xml(xml,
                                                         "signal_lr_lnpdf")
        self.zero_lag_lr_lnpdf = rate.BinnedLnPDF.from_xml(
            xml, "zero_lag_lr_lnpdf")
        self.segments = ligolw_param.get_pyvalue(xml, u"segments").strip()
        self.segments = \
            segments_utils.from_range_strings(self.segments.split(",")
                                              if self.segments else [], float)

        return self

    def new_with_extinction(self, extinct_zerowise_elems):
        """
        Adapted from far.py, gstlal-inspiral-1.6 branch
        """
        bg = self.noise_lr_lnpdf.array
        x = self.noise_lr_lnpdf.bins[0].centres()
        assert (x == self.zero_lag_lr_lnpdf.bins[0].centres()).all()
        ccdf = bg[::-1].cumsum()[::-1]
        ccdf /= ccdf[0]

        def mk_survival_probability(rate_eff, m):
            return np.exp(-rate_eff * ccdf**m)

        zl = self.zero_lag_lr_lnpdf.copy()
        zl.array[:extinct_zerowise_elems] = 0.
        if not zl.array.any():
            raise ValueError("zero-lag counts are all zero")

        mode, = zl.argmax()
        zlcumsum = zl.array.cumsum()
        assert zlcumsum[-1] > 1000, \
            "Need at least 1000 zero lag events to compute extinction model"
        ten_thousand_events_lr = \
            x[zlcumsum.searchsorted(zlcumsum[-1] - 10000)]
        # Adjust the mode to be at 10,000 events if that LR is higher
        if ten_thousand_events_lr > mode:
            mode = ten_thousand_events_lr
        one_hundred_events_lr = x[zlcumsum.searchsorted(zlcumsum[-1] - 100)]
        mask = (x < mode) | (x > one_hundred_events_lr)
        zl = np.ma.masked_array(zl.array, mask)
        bg = np.ma.masked_array(bg, mask)

        def ssr(params):
            norm, rate_eff, m = params
            if norm <= 0. or rate_eff <= 0. or m <= 0.:
                return np.inf

            model = norm * bg * mk_survival_probability(rate_eff, m)
            model_var = np.where(model > 1., model, 1.)
            square_error = (zl - model)**2. / model_var

            return np.trapz(square_error, x)

        norm, rate_eff, m = optimize.fmin(ssr,
                                          (zl.sum() / bg.sum(), zl.sum(), 5.),
                                          xtol=1e-8, ftol=1e-8, disp=0)

        # compute survival probability model from best fit
        survival_probability = mk_survival_probability(rate_eff, m)

        # apply to background counts and signal counts
        self.noise_lr_lnpdf.array *= survival_probability
        self.noise_lr_lnpdf.normalize()
        self.signal_lr_lnpdf.array *= survival_probability
        self.signal_lr_lnpdf.normalize()

        return self

class FAPFAR(object):
    """
    Adapted from far.py, gstlal-inspiral-1.6 branch
    """
    def __init__(self, rankingstatpdf, extinct_zerowise_elems):
        # input checks
        if not rankingstatpdf.zero_lag_lr_lnpdf.array.any():
            raise ValueError("RankingStatPDF's zero-lag counts are all zero")

        # save the livetime
        self.livetime = float(abs(rankingstatpdf.segments))

        zl = rankingstatpdf.zero_lag_lr_lnpdf.copy()
        zl.array[:extinct_zerowise_elems] = 0.
        rate_normalization_lr_threshold, = zl.argmax()

        # record trials factor, with safety checks
        counts = rankingstatpdf.zero_lag_lr_lnpdf.count
        assert not np.isnan(counts.array).any(), \
            "zero lag log likelihood ratio counts contain NaNs"
        assert (counts.array >= 0.).all(), \
            "zero lag log likelihood ratio rates contain negative values"
        self.count_above_threshold = \
            counts[rate_normalization_lr_threshold:, ].sum()

        # get noise model ranking stat values and event counts from
        # bins
        noise_lnpdf = rankingstatpdf.noise_lr_lnpdf

        threshold_index = \
            noise_lnpdf.bins[0][rate_normalization_lr_threshold]
        ranks = \
            noise_lnpdf.bins[0].lower()[threshold_index:]
        counts = noise_lnpdf.array[threshold_index:]
        assert not np.isnan(counts).any(), \
            "background log likelihood ratio rates contain NaNs"
        assert (counts >= 0.).all(), \
            "background log likelihood ratio rates contain negative values"

        ccdf = counts[::-1].cumsum()[::-1]
        ccdf /= ccdf[0]
        ccdf = poisson_p_not_0(ccdf)

        # safety checks
        assert not np.isnan(ranks).any(), \
            "log likelihood ratio co-ordinates contain NaNs"
        assert not np.isinf(ranks).any(), \
            "log likelihood ratio co-ordinates are not all finite"
        assert not np.isnan(ccdf).any(), \
            "log likelihood ratio CCDF contains NaNs"
        assert ((0. <= ccdf) & (ccdf <= 1.)).all(), \
            "log likelihood ratio CCDF failed to be normalized"

        # build interpolator.
        self.ccdf_interpolator = interpolate.interp1d(ranks, ccdf)

        self.minrank = ranks[0]
        self.maxrank = ranks[-1]

    @assert_probability
    def ccdf_from_rank(self, rank):
        return self.ccdf_interpolator(np.clip(rank,
                                              self.minrank,
                                              self.maxrank))

    def far_from_rank(self, rank):
        log_tdp = np.log1p(-self.ccdf_from_rank(rank))
        return self.count_above_threshold * -log_tdp / self.livetime

def _parse_likelihood_control_doc(xmldoc):
    name = "gstlal_inspiral_likelihood"
    rankingstatpdf = _RankingStatPDF.from_xml(xmldoc, name)
    if rankingstatpdf is None:
        raise ValueError("document does not contain likelihood ratio data")
    return rankingstatpdf


@ligolw_array.use_in
@ligolw_param.use_in
@lsctables.use_in
class _ContentHandler(LIGOLWContentHandler):
    pass

def _get_ln_f_over_b(ranking_data_bytes,
                     ln_likelihood_ratios,
                     livetime,
                     extinct_zerowise_elems):

    ranking_data_xmldoc = ligolw_utils.load_fileobj(
        io.BytesIO(ranking_data_bytes), contenthandler=_ContentHandler)
    rankingstatpdf = _parse_likelihood_control_doc(ranking_data_xmldoc)
    # affect the zeroing of the PDFs below threshold by hacking the
    # histograms. Do the indexing ourselves to not 0 the bin @ threshold
    noise_lr_lnpdf = rankingstatpdf.noise_lr_lnpdf
    signal_lr_lnpdf = rankingstatpdf.signal_lr_lnpdf
    zero_lag_lr_lnpdf = rankingstatpdf.zero_lag_lr_lnpdf
    ssorted = zero_lag_lr_lnpdf.array.cumsum()[-1] - 10000
    idx = zero_lag_lr_lnpdf.array.cumsum().searchsorted(ssorted)

    ln_likelihood_ratio_threshold = \
        zero_lag_lr_lnpdf.bins[0].lower()[idx]

    # Compute FAR for threshold, and estimate
    # terrestrial count consistent with threshold
    fapfar = \
        FAPFAR(rankingstatpdf.new_with_extinction(extinct_zerowise_elems),
               extinct_zerowise_elems)
    far_threshold = fapfar.far_from_rank(ln_likelihood_ratio_threshold)
    lam_0 = far_threshold*livetime

    rankingstatpdf.noise_lr_lnpdf.array[
        :noise_lr_lnpdf.bins[0][ln_likelihood_ratio_threshold]] \
        = 0.
    rankingstatpdf.noise_lr_lnpdf.normalize()

    rankingstatpdf.signal_lr_lnpdf.array[
        :signal_lr_lnpdf.bins[0][ln_likelihood_ratio_threshold]] \
        = 0.
    rankingstatpdf.signal_lr_lnpdf.normalize()

    rankingstatpdf.zero_lag_lr_lnpdf.array[
        :zero_lag_lr_lnpdf.bins[0][ln_likelihood_ratio_threshold]] \
        = 0.
    rankingstatpdf.zero_lag_lr_lnpdf.normalize()

    f = rankingstatpdf.signal_lr_lnpdf
    b = rankingstatpdf.noise_lr_lnpdf
    ln_f_over_b = \
        np.array([f[ln_lr, ] - b[ln_lr, ] for ln_lr in ln_likelihood_ratios])
    if np.isnan(ln_f_over_b).any():
        raise ValueError("NaN encountered in ranking statistic PDF ratios")
    if np.isinf(np.exp(ln_f_over_b)).any():
        raise ValueError(
            "infinity encountered in ranking statistic PDF ratios")
    return ln_f_over_b, lam_0

def _get_event_ln_likelihood_ratio_svd_endtime_mass(coinc_bytes):
    coinc_xmldoc = ligolw_utils.load_fileobj(
        io.BytesIO(coinc_bytes), contenthandler=_ContentHandler)
    coinc_event, = lsctables.CoincTable.get_table(coinc_xmldoc)
    coinc_inspiral, = lsctables.CoincInspiralTable.get_table(coinc_xmldoc)
    sngl_inspiral = lsctables.SnglInspiralTable.get_table(coinc_xmldoc)

    assert all([sngl_inspiral[i].Gamma0 == sngl_inspiral[i+1].Gamma0
                for i in range(len(sngl_inspiral)-1)]), \
        "svd bank different between ifos!"
    return (coinc_event.likelihood,
            sngl_inspiral[0].mass1,
            sngl_inspiral[0].mass2,
            sngl_inspiral[0].spin1z,
            sngl_inspiral[0].spin2z,
            coinc_inspiral.snr,
            coinc_inspiral.combined_far)



from importlib import resources

import numpy as np
import pytest

from .. import p_astro
from . import data


def source_type(label, w_fgmc):
    return p_astro.SourceType(label=label, w_fgmc=w_fgmc)

def counts_instance_func(f_divby_b,prior_type, terr_source_instance,
                         fix_sources_dict,**source_instances):
    return p_astro.MarginalizedPosterior(f_divby_b=f_divby_b,
                                         prior_type=prior_type,
                                         terr_source=terr_source_instance,
                                         fix_sources=fix_sources_dict,
                                         **source_instances)

def compute_pastro(category, idx, counts_instance):
    return counts_instance.pastro(categories=[category], trigger_idx=idx)[0]

@pytest.mark.filterwarnings('ignore::RuntimeWarning')
def test_p_astro_categories():
    with resources.open_binary(data, 'zerolag_svd_bank_nums.txt') as f:
        zerolag_lnls, svd_bank_nums = np.loadtxt(f, unpack=True)
    with resources.open_binary(data, 'f_divby_b.txt') as f:
        fb = np.loadtxt(f, unpack=True)

    # Acquire activation counts for each source category
    with resources.open_binary(data, 'bns_wellfound_hits.txt') as f:
        a_bns = np.genfromtxt(f, names=True)["hit_count"]
    with resources.open_binary(data, 'nsbh_wellfound_hits.txt') as f:
        a_nsbh = np.genfromtxt(f, names=True)["hit_count"]
    with resources.open_binary(data, 'bbh_wellfound_hits.txt') as f:
        a_bbh = np.genfromtxt(f, names=True)["hit_count"]
    with resources.open_binary(data, 'background_hits.txt') as f:
        a_terr = np.genfromtxt(f, names=True)["hit_count"]

    # Normalize activation counts
    a_hat_bns = a_bns/(np.sum(a_bns))
    a_hat_nsbh = a_nsbh/(np.sum(a_nsbh))
    a_hat_bbh = a_bbh/(np.sum(a_bbh))
    a_hat_terr = a_terr/(np.sum(a_terr))

    # Assign one weight per source category to each candidate event
    w_bns = np.array([a_hat_bns[int(svd)] for svd in svd_bank_nums])
    w_nsbh = np.array([a_hat_nsbh[int(svd)] for svd in svd_bank_nums])
    w_bbh = np.array([a_hat_bbh[int(svd)] for svd in svd_bank_nums])
    #w_terr = np.ones(len(w_bns))
    w_terr = np.array([a_hat_terr[int(svd)] for svd in svd_bank_nums])

    N = len(zerolag_lnls) # number of candidate events
    counts_instance = counts_instance_func(
        f_divby_b=fb,
        prior_type="Uniform",
        terr_source_instance=source_type(label="Terr",w_fgmc=w_terr),
        fix_sources_dict={"Terr":N},
        bns_inst=source_type(label="BNS",w_fgmc=w_bns),
        bbh_inst=source_type(label="BBH",w_fgmc=w_bbh),
        nsbh_inst=source_type(label="NSBH",w_fgmc=w_nsbh))

    p_astro_bns = compute_pastro(category="BNS", idx=N-1,
                                 counts_instance=counts_instance)

    p_astro_bbh = compute_pastro(category="BBH", idx=N-1,
                                 counts_instance=counts_instance)

    p_astro_nsbh = compute_pastro(category="NSBH", idx=N-1,
                                 counts_instance=counts_instance)

    assert np.isclose(p_astro_bns, 0.)
    assert np.isclose(p_astro_bbh, 0.446, rtol=1e-03)
    assert np.isclose(p_astro_nsbh, 0.)



from importlib import resources
import json
import numpy as np
import pytest

from .. import p_astro_gstlal_utils as gstlal
from .. import p_astro_computation as pastrocomp
from . import data as test_data
from .. import data as pkg_data


@pytest.fixture
def coinc_bytes_1():
    return resources.read_binary(test_data, 'coinc_G322589.xml')


@pytest.fixture
def coinc_bytes_2():
    return resources.read_binary(test_data, 'coinc_G5351.xml')


@pytest.fixture
def coinc_bytes_3():
    return resources.read_binary(test_data, 'coinc.xml')


@pytest.fixture
def ranking_data_bytes():
    return resources.read_binary(test_data, 'ranking_data_G322589.xml.gz')


@pytest.fixture
def activation_counts_bytes():
    return resources.read_binary(
        pkg_data, 'H1L1V1-weights-bins_686-1126051217-61603201.json')

@pytest.fixture
def mean_values_bytes():
    return resources.read_binary(
        pkg_data, 'H1L1V1-mean_counts-1126051217-61603201.json')

def test_get_ln_f_over_b(ranking_data_bytes):
    """Test to check `_get_ln_f_over_b` returns values
    which are not inf or nan for high ln_likelihood_ratio.
    Test to check the computation of terrestrial count
    `lam_0` returns the expected value."""
    ln_f_over_b, lam_0 = gstlal._get_ln_f_over_b(
        ranking_data_bytes,
        [100., 200., 300.],
        livetime=14394240,
        extinct_zerowise_elems=40)
    assert np.all(np.isfinite(ln_f_over_b))
    assert lam_0 == pytest.approx(32575, abs=1.0)


def test_get_event_ln_likelihood_ratio_svd_endtime_mass(coinc_bytes_1):
    likelihood, mass1, mass2, spin1z, spin2z, snr, far = \
        gstlal._get_event_ln_likelihood_ratio_svd_endtime_mass(
            coinc_bytes_1)
    assert mass1 == pytest.approx(2.8, abs=0.1)
    assert mass2 == pytest.approx(1.0, abs=0.1)
    assert spin1z == pytest.approx(-0.99, abs=0.01)
    assert spin2z == pytest.approx(0.049, abs=0.01)
    assert likelihood == pytest.approx(21.65, abs=0.1)


def test_compute_p_astro_1(coinc_bytes_1,
                           ranking_data_bytes,
                           mean_values_bytes,
                           activation_counts_bytes):
    """Test to call `compute_p_astro` on gracedb event G322589.
    m1 = 2.7, m2 = 1.0 solar mass for this event"""
    files = coinc_bytes_1, ranking_data_bytes

    activation_counts_dict = json.loads(activation_counts_bytes)
    mean_values_dict = json.loads(mean_values_bytes)

    likelihood, mass1, mass2, spin1z, spin2z, snr, far = \
        gstlal._get_event_ln_likelihood_ratio_svd_endtime_mass(
            coinc_bytes_1)

    ln_f_over_b, lam_0 = gstlal._get_ln_f_over_b(ranking_data_bytes,
                                                 [likelihood],
                                                 livetime=14394240,
                                                 extinct_zerowise_elems=40)

    mean_values_dict["counts_Terrestrial"] = lam_0
    num_bins = 686

    p_astros = \
        pastrocomp.evaluate_p_astro_from_bayesfac(np.exp(ln_f_over_b),
                                                  mean_values_dict,
                                                  mass1, mass2,
                                                  spin1z, spin2z,
                                                  num_bins,
                                                  activation_counts_dict)

    assert p_astros['BNS'] == pytest.approx(1, abs=1e-2)
    assert p_astros['NSBH'] == pytest.approx(0, abs=1e-2)
    assert p_astros['BBH'] == pytest.approx(0, abs=1e-2)
    assert p_astros['Terrestrial'] == pytest.approx(0, abs=1e-2)


def test_compute_p_astro_2(coinc_bytes_2,
                           ranking_data_bytes,
                           mean_values_bytes,
                           activation_counts_bytes):
    """Test to call `compute_p_astro` on gracedb event G5351. Use ranking
    data from an O3 event G330299. m1 = 1.1, m2 = 1.0 solar mass
    for this event. FAR = 1.9e-6, P_terr has a moderate value."""
    files = coinc_bytes_2, ranking_data_bytes

    activation_counts_dict = json.loads(activation_counts_bytes)
    mean_values_dict = json.loads(mean_values_bytes)

    likelihood, mass1, mass2, spin1z, spin2z, snr, far = \
        gstlal._get_event_ln_likelihood_ratio_svd_endtime_mass(
            coinc_bytes_2)

    ln_f_over_b, lam_0 = gstlal._get_ln_f_over_b(ranking_data_bytes,
                                                 [likelihood],
                                                 livetime=14394240,
                                                 extinct_zerowise_elems=40)

    mean_values_dict["counts_Terrestrial"] = lam_0
    num_bins = 686

    p_astros = \
        pastrocomp.evaluate_p_astro_from_bayesfac(np.exp(ln_f_over_b),
                                                  mean_values_dict,
                                                  mass1, mass2,
                                                  spin1z, spin2z,
                                                  num_bins,
                                                  activation_counts_dict)

    assert p_astros['Terrestrial'] > 0.25



import numpy as np

def p_astro_update(category, event_bayesfac_dict, mean_values_dict):
    """
    Compute `p_astro` for a new event using mean values of Poisson expected
    counts constructed from all the previous events. Invoked with every new
    GraceDB entry.

    Parameters
    ----------
    category : string
        source category
    event_bayesfac_dict : dictionary
        event Bayes factors
    mean_values_dict : dictionary
        mean values of Poisson counts

    Returns
    -------
    p_astro : float
        p_astro by source category
    """
    if category == "counts_Terrestrial":
        numerator = mean_values_dict["counts_Terrestrial"]
    else:
        numerator = \
            event_bayesfac_dict[category] * mean_values_dict[category]

    denominator = mean_values_dict["counts_Terrestrial"] + \
        np.sum([mean_values_dict[key] * event_bayesfac_dict[key]
                for key in event_bayesfac_dict.keys()])

    return numerator / denominator


def evaluate_p_astro_from_bayesfac(astro_bayesfac,
                                   mean_values_dict,
                                   mass1,
                                   mass2,
                                   spin1z=None,
                                   spin2z=None,
                                   num_bins=None,
                                   activation_counts=None):
    """
    Evaluates `p_astro` for a new event using Bayes factor, masses, and number
    of astrophysical categories. Invoked with every new GraceDB entry.

    Parameters
    ----------
    astro_bayesfac : float
        astrophysical Bayes factor
    mean_values_dict: dictionary
        mean values of Poisson counts
    mass1 : float
        event mass1
    mass2 : float
        event mass2
    spin1z : float
        event spin1z
    spin2z : float
        event spin2z
    url_weights_key: str
        url config key pointing to weights file

    Returns
    -------
    p_astro : dictionary
        p_astro for all source categories
    """

    a_hat_bns, a_hat_bbh, a_hat_nsbh, num_bins = \
        make_weights_from_histograms(mass1,
                                     mass2,
                                     spin1z,
                                     spin2z,
                                     num_bins,
                                     activation_counts)

    # Compute category-wise Bayes factors
    # from astrophysical Bayes factor
    rescaled_fb = num_bins * astro_bayesfac
    bns_bayesfac = a_hat_bns * rescaled_fb
    nsbh_bayesfac = a_hat_nsbh * rescaled_fb
    bbh_bayesfac = a_hat_bbh * rescaled_fb

    # Construct category-wise Bayes factor dictionary
    event_bayesfac_dict = {"counts_BNS": bns_bayesfac,
                           "counts_NSBH": nsbh_bayesfac,
                           "counts_BBH": bbh_bayesfac}

    # Compute the p-astro values for each source category
    # using the mean values
    p_astro_values = {}
    for category in mean_values_dict:
        p_astro_values[category.split("_")[1]] = \
            p_astro_update(category=category,
                           event_bayesfac_dict=event_bayesfac_dict,
                           mean_values_dict=mean_values_dict)

    return p_astro_values


def make_weights_from_hardcuts(mass1, mass2):
    """
    Construct binary weights from component masses based on cuts in component
    mass space that define astrophysical source categories. To be used for
    MBTA, PyCBC and SPIIR.

    Parameters
    ----------
    mass1 : float
        heavier component mass of the event
    mass2 : float
        lighter component mass of the event

    Returns
    -------
    a_bns, a_bbh, a_nshb, a_mg : floats
        binary weights (i.e, 1 or 0)
    """

    a_hat_bns = int(mass1 <= 3 and mass2 <= 3)
    a_hat_bbh = int(mass1 > 5 and mass2 > 5)
    a_hat_nsbh = int(min(mass1, mass2) <= 3 and
                     max(mass1, mass2) > 3)
    num_bins = 3

    return a_hat_bns, a_hat_bbh, a_hat_nsbh, num_bins


def closest_template(params, params_list):
    """
    Associate event's template to a template in the template bank. The assumed
    bank is the one used by Gstlal. Hence, for Gstlal events, the association
    should be exact, up to rounding errors.

    Parameters
    ----------
    params : tuple of floats
        intrinsic params of event template
    params_list: list of strings
        list of template bank's template params

    Returns
    -------
    key : string
        params of template in template bank
        matching event's template
    """
    params_array = np.array(list(map(eval, params_list)))
    idx = np.argmin(np.sum((params_array-params)**2, axis=1))
    num_params = len(params_array[idx])
    template = params_array[idx]
    string = '(' + ', '.join(['{:3.8f}']*num_params) + ')'
    key = string.format(*template)
    return key


def make_weights_from_histograms(mass1,
                                 mass2,
                                 spin1z,
                                 spin2z,
                                 num_bins=None,
                                 activation_counts=None):
    """
    Construct binary weights from bin number provided by GstLAL, and a weights
    matrix pre-constructed and stored in a file, to be read from a url. The
    weights are keyed on template parameters of Gstlal's template bank. If that
    doesn't work, construct binary weights.

    Parameters
    ----------
    mass1 : float
        heavier component mass of the event
    mass2 : float
        lighter component mass of the event
    spin1z : float
        z component spin of heavier mass
    spin2z : float
        z component spin of lighter mass

    Returns
    -------
    a_hat_bns, a_hat_bbh, a_hat_nsbh, a_hat_mg : floats
        mass-based template weights
    """

    if activation_counts is None or num_bins is None:
        a_hat_bns, a_hat_bbh, a_hat_nsbh, num_bins = \
            make_weights_from_hardcuts(mass1, mass2)
    else:
        params = (mass1, mass2, spin1z, spin2z)
        params_list = list(activation_counts.keys())
        key = closest_template(params, params_list)
        event_weights_dict = activation_counts[key]
        source_types = np.sort(list(event_weights_dict.keys()))
        a_hat_bbh, a_hat_bns, a_hat_nsbh = \
            tuple([event_weights_dict[s] for s in source_types])

    return a_hat_bns, a_hat_bbh, a_hat_nsbh, num_bins


def choose_snr(far, snr, pipeline, instruments, threshold_dict):
    """
    Given a pipeline and combination of instruments, return an SNR that does
    not exceed the SNR threshold for FARs below a FAR threshold. The SNR and
    FAR thresholds are read from a file containing these values keyed on
    pipelines and instrument combinations.

    Parameters
    ----------
    far : float
        false alarm rate of the event
    snr : float
        SNR of the event
    pipeline : string
        pipeline that posted the event
    instruments : set
        set of instruments that detected the event
    threshold_dict : dictionary
        dictionary of FAR-SNR thresholds for instrument
        combinations

    Returns
    -------
    snr : float
        limiting SNR value
    """

    if pipeline == "gstlal":
        snr_choice = snr
    else:
        inst_sorted = ",".join(sorted(instruments))
        far_t = threshold_dict[pipeline][inst_sorted]["far"]
        snr_t = threshold_dict[pipeline][inst_sorted]["snr"]
        if far < far_t and snr > snr_t:
            snr_choice = snr_t
        else:
            snr_choice = snr

    return snr_choice


def get_f_over_b(far, snr_choice, far_star, snr_star):
    """
    Compute bayesfactor for non-gstlal pipelines using an
    approximate ("Toy") model.
    
    Parameters
    ----------
    far : float
        false alarm rate of the event
    snr : float
        SNR of the event
    far_star : float
        threshold false alarm rate
    snr_star : float
        threshold SNR

    Returns
    -------
    bayesfactor : float
        bayesfactor of event
    """
    # Compute astrophysical bayesfactor for
    # GraceDB event
    fground = 3 * snr_star**3 / (snr_choice**4)
    bground = far / far_star
    return fground / bground



from importlib import resources
import json

import pytest

from .. import p_astro_computation as pastrocomp
from .. import data


@pytest.fixture
def thresholds_bytes():
    return resources.read_binary(
        data, 'H1L1V1-pipeline-far_snr-thresholds.json')


@pytest.fixture
def mean_counts_bytes():
    return resources.read_binary(
        data, 'H1L1V1-mean_counts-1126051217-61603201.json')


@pytest.mark.parametrize(
    'far,pipeline,instruments,snr_thresh,val',
    ([1e-25, 'mbta', {'H1', 'L1', 'V1'}, 12, 1],
     [1e-8, 'mbta', {'L1', 'V1'}, 33, 0.08],
     [6e-10, 'mbta', {'H1', 'V1'}, 10, 0.99],
     [7.6e-59, 'spiir', {'H1', 'L1', 'V1'}, 33, 1],
     [1e-10, 'pycbc', {'H1', 'L1'}, 10, 1],
     [1e-10, 'pycbc', {'H1', 'L1', 'V1'}, 10, 1]))
def test_compute_p_astro_bns(far, pipeline, instruments,
                             snr_thresh, val,
                             thresholds_bytes,
                             mean_counts_bytes):
    """Test p_astro values using CBC catalog paper
    values for GW170817, for various mock-FARs, to test
    handling of this very loud event for MBTA, PyCBC
    and spiir.
    """
    # values based on G322759
    snr = 33.
    mass1 = 1.77
    mass2 = 1.07

    far_star = 1. / (30. * 86400.)
    snr_star = 8.5
    livetime = 166.6 * 86400.

    thresholds_dict = json.loads(thresholds_bytes)
    mean_values_dict = json.loads(mean_counts_bytes)
    mean_values_dict["counts_Terrestrial"] = far_star * livetime

    snr_choice = pastrocomp.choose_snr(far,
                                       snr,
                                       pipeline,
                                       instruments,
                                       thresholds_dict)

    astro_bayesfac = pastrocomp.get_f_over_b(far,
                                             snr_choice,
                                             far_star,
                                             snr_star)

    p_astros = \
        pastrocomp.evaluate_p_astro_from_bayesfac(astro_bayesfac,
                                                  mean_values_dict,
                                                  mass1, mass2)

    assert pytest.approx(snr_thresh, abs=1e-1) == snr_choice
    assert pytest.approx(p_astros['BNS'], abs=1e-1) == val


@pytest.mark.parametrize(
    'pipeline,instruments,far,snr,snr_c',
    (['mbta', {'H1', 'L1'}, 4e-10, 50, 50],
     ['mbta', {'H1', 'L1'}, 2e-10, 50, 10],
     ['mbta', {'L1', 'V1'}, 8e-10, 50, 50],
     ['mbta', {'L1', 'V1'}, 6e-10, 50, 10],
     ['mbta', {'H1', 'V1'}, 8e-10, 50, 50],
     ['mbta', {'H1', 'V1'}, 6e-10, 50, 10],
     ['mbta', {'H1', 'L1', 'V1'}, 1e-13, 50, 50],
     ['mbta', {'H1', 'L1', 'V1'}, 1e-15, 50, 12],
     ['pycbc', {'H1', 'L1'}, 4e-10, 50, 50],
     ['pycbc', {'H1', 'L1'}, 2e-10, 50, 10],
     ['spiir', {'H1', 'L1'}, 4e-20, 50, 50],
     ['gstlal', None, 4e-20, 50, 50]))
def test_compute_choose_snr(pipeline, instruments, far,
                            snr, snr_c,thresholds_bytes):
    """For various mock-FARs, test the snr returned for
       very loud MBTA, PyCBC and spiir events.
    """

    thresholds_dict = json.loads(thresholds_bytes)
    snr_choice = pastrocomp.choose_snr(far, snr, pipeline, instruments, thresholds_dict)

    assert pytest.approx(snr_choice, abs=1e-2) == snr_c



#!/usr/bin/python
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


from setuptools import setup, find_packages

setup(
    name='p_astro',
    version='1.0.1',
    url='https://lscsoft.docs.ligo.org/p-astro/',
    author='Shasvath Kapadia, Deep Chatterjee, Shaon Ghosh',
    author_email='shasvath.kapadia@ligo.org, deep.chatterjee@ligo.org, shaon.ghosh@ligo.org',
    maintainer="Deep Chatterjee",
    maintainer_email="deep.chatterjee@ligo.org",
    description='Low-latency classification of GW triggers from compact binary coalescence',
    license='GNU General Public License Version 3',
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Topic :: Internet",
        "Topic :: Scientific/Engineering :: Astronomy",
        "Topic :: Scientific/Engineering :: Physics"
    ],
    packages=find_packages(),
    include_package_data=True,
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    namespace_packages=['ligo'],
    install_requires=[
        'astropy',
        'lalsuite',
        'numpy',
        'python-ligo-lw',
        'scipy',
        'h5py'
    ],
    python_requires='>=3.7',
    entry_points = {
        'console_scripts': [
            'p_astro_histogram_by_bin=ligo.p_astro_utils:histogram_by_bin',
            'p_astro_compute_means=ligo.p_astro_utils:compute_counts_mean'
        ]
    }
)



# Copyright (C) 2018 Shasvath Kapadia, Deep Chatterjee
#                    Heather Fong, Surabhi Sachdev
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import io
import json
import sqlite3

from argparse import ArgumentParser
from configparser import ConfigParser
import numpy as np

from astropy.table import Table

from ligo.lw import ligolw
from ligo.lw.ligolw import LIGOLWContentHandler
from ligo.lw import array as ligolw_array
from ligo.lw import param as ligolw_param
from ligo.lw import dbtables
from ligo.lw import utils as ligolw_utils
from ligo.lw import lsctables
from lalinspiral import thinca
from lal import rate

from .p_astro import SourceType, MarginalizedPosterior


class _RankingStatPDF(object):
    ligo_lw_name_suffix = "gstlal_inspiral_rankingstatpdf"

    @classmethod
    def from_xml(cls, xml, name):
        """
        Find the root of the XML tree containing the
        serialization of this object
        """
        xml, = [elem for elem in
                xml.getElementsByTagName(ligolw.LIGO_LW.tagName)
                if elem.hasAttribute("Name") and
                elem.Name == "%s:%s" % (name, cls.ligo_lw_name_suffix)]
        # create a uninitialized instance
        self = super().__new__(cls)
        # populate from XML
        self.noise_lr_lnpdf = rate.BinnedLnPDF.from_xml(xml, "noise_lr_lnpdf")
        self.signal_lr_lnpdf = rate.BinnedLnPDF.from_xml(xml,
                                                         "signal_lr_lnpdf")
        self.zero_lag_lr_lnpdf = rate.BinnedLnPDF.from_xml(
            xml, "zero_lag_lr_lnpdf")
        return self


def _parse_likelihood_control_doc(xmldoc):
    name = "gstlal_inspiral_likelihood"
    rankingstatpdf = _RankingStatPDF.from_xml(xmldoc, name)
    if rankingstatpdf is None:
        raise ValueError("document does not contain likelihood ratio data")
    return rankingstatpdf


@ligolw_array.use_in
@ligolw_param.use_in
@lsctables.use_in
class _ContentHandler(LIGOLWContentHandler):
    pass

def _get_ln_f_over_b(ranking_data_bytes,
                     ln_likelihood_ratios,
                     ln_likelihood_ratio_threshold):
    ranking_data_xmldoc, _ = ligolw_utils.load_fileobj(
        io.BytesIO(ranking_data_bytes), contenthandler=_ContentHandler)
    rankingstatpdf = _parse_likelihood_control_doc(ranking_data_xmldoc)
    # affect the zeroing of the PDFs below threshold by hacking the
    # histograms. Do the indexing ourselves to not 0 the bin @ threshold
    noise_lr_lnpdf = rankingstatpdf.noise_lr_lnpdf
    rankingstatpdf.noise_lr_lnpdf.array[
        :noise_lr_lnpdf.bins[0][ln_likelihood_ratio_threshold]] \
        = 0.
    rankingstatpdf.noise_lr_lnpdf.normalize()
    signal_lr_lnpdf = rankingstatpdf.signal_lr_lnpdf
    rankingstatpdf.signal_lr_lnpdf.array[
        :signal_lr_lnpdf.bins[0][ln_likelihood_ratio_threshold]] \
        = 0.
    rankingstatpdf.signal_lr_lnpdf.normalize()
    zero_lag_lr_lnpdf = rankingstatpdf.zero_lag_lr_lnpdf
    rankingstatpdf.zero_lag_lr_lnpdf.array[
        :zero_lag_lr_lnpdf.bins[0][ln_likelihood_ratio_threshold]] \
        = 0.
    rankingstatpdf.zero_lag_lr_lnpdf.normalize()

    f = rankingstatpdf.signal_lr_lnpdf
    b = rankingstatpdf.noise_lr_lnpdf
    ln_f_over_b = \
        np.array([f[ln_lr, ] - b[ln_lr, ] for ln_lr in ln_likelihood_ratios])
    if np.isnan(ln_f_over_b).any():
        raise ValueError("NaN encountered in ranking statistic PDF ratios")
    if np.isinf(np.exp(ln_f_over_b)).any():
        raise ValueError(
            "infinity encountered in ranking statistic PDF ratios")
    return ln_f_over_b


def _load_search_results(conn, end_time, mass, ln_likelihood_ratio_threshold):
    """Queries SQLite file for background trigger data.
    The query has an extra check to make sure the event is not
    double counted.
    """
    cur = conn.cursor()

    xmldoc = dbtables.get_xml(conn)
    definer_id = lsctables.CoincDefTable.get_table(xmldoc).get_coinc_def_id(
        thinca.InspiralCoincDef.search,
        thinca.InspiralCoincDef.search_coinc_type,
        create_new=False)
    cur.execute("""
SELECT
        coinc_event.likelihood,
        (
        SELECT sngl_inspiral.Gamma1
        FROM
                sngl_inspiral JOIN coinc_event_map ON (
                        coinc_event_map.table_name == "sngl_inspiral"
                        AND coinc_event_map.event_id == sngl_inspiral.event_id
                )
        WHERE
        coinc_event_map.coinc_event_id == coinc_inspiral.coinc_event_id
        LIMIT 1
        ),
        EXISTS (
    SELECT
*
    FROM
time_slide
    WHERE
time_slide.time_slide_id == coinc_event.time_slide_id
AND time_slide.offset != 0
        )
FROM
        coinc_event JOIN coinc_inspiral ON (
    coinc_inspiral.coinc_event_id == coinc_event.coinc_event_id
        )
WHERE
        coinc_event.coinc_def_id == ?
        AND coinc_event.likelihood >= ?
        AND coinc_inspiral.end_time != ?
        AND coinc_inspiral.mass != ?""", (definer_id,
                                          ln_likelihood_ratio_threshold,
                                          end_time,
                                          mass))
    ln_likelihood_ratio, svd_banks, is_background = np.array(cur.fetchall()).T
    background_ln_likelihood_ratios = \
        ln_likelihood_ratio[is_background.astype(bool)]
    zerolag_ln_likelihood_ratios = \
        ln_likelihood_ratio[np.logical_not(is_background.astype(bool))]
    svd_banks = svd_banks[np.logical_not(is_background.astype(bool))].astype(int)

    return background_ln_likelihood_ratios, \
        zerolag_ln_likelihood_ratios, svd_banks


def _get_counts_instance(ln_f_over_b,
                         svd_bank_nums,
                         num_svd_bins,
                         activation_counts_table,
                         lnl_threshold=None,
                         fgmc_data_history=None,
                         prior_type="Uniform"):

    a_hat_bns = np.array(activation_counts_table['bns'], dtype=float)
    a_hat_bns /= np.sum(a_hat_bns)
    a_hat_bbh = np.array(activation_counts_table['bbh'], dtype=float)
    a_hat_bbh /= np.sum(a_hat_bbh)
    a_hat_nsbh = np.array(activation_counts_table['nsbh'], dtype=float)
    a_hat_nsbh /= np.sum(a_hat_nsbh)
    a_hat_mg = np.array(activation_counts_table['mg'], dtype=float)
    a_hat_mg /= np.sum(a_hat_mg)
    

    w_bns = num_svd_bins*np.take(a_hat_bns, svd_bank_nums)
    w_nsbh = num_svd_bins*np.take(a_hat_nsbh, svd_bank_nums)
    w_bbh = num_svd_bins*np.take(a_hat_bbh, svd_bank_nums)
    w_mg = num_svd_bins*np.take(a_hat_mg, svd_bank_nums)

    if fgmc_data_history is not None and lnl_threshold is not None:

        history = np.genfromtxt(fgmc_data_history,names=True)
        idx = history["lnl"] >= lnl_threshold

        w_bns = np.array(list(num_svd_bins*history["w_bns"][idx])+list(w_bns))
        w_nsbh = np.array(list(num_svd_bins*history["w_nsbh"][idx])+list(w_nsbh))
        w_bbh = np.array(list(num_svd_bins*history["w_bbh"][idx])+list(w_bbh))
        w_mg = np.array(list(num_svd_bins*history["w_mg"][idx])+list(w_mg))

        ln_f_over_b = np.array(list(history["ln_f_over_b"][idx]) +
                               list(ln_f_over_b))

    num_f_over_b = len(ln_f_over_b)
    w_terr = np.ones(num_f_over_b)
    fb = np.exp(ln_f_over_b)
    return MarginalizedPosterior(f_divby_b=fb, prior_type=prior_type,
                                 terr_source=SourceType(label="counts_Terrestrial",
                                                        w_fgmc=w_terr),
                                 fix_sources={"counts_Terrestrial": num_f_over_b},
                                 bns_inst=SourceType(label="counts_BNS",
                                                     w_fgmc=w_bns),
                                 bbh_inst=SourceType(label="counts_BBH",
                                                     w_fgmc=w_bbh),
                                 nsbh_inst=SourceType(label="counts_NSBH",
                                                      w_fgmc=w_nsbh),
                                 mg_inst=SourceType(label="counts_MassGap",
                                                      w_fgmc=w_mg),
                                 verbose=False)


def compute_counts_mean():
    parser = ArgumentParser("Compute mean on Poisson count")
    parser.add_argument(
        '-i', '--input', required=True,
        help='Astropy table storing activation counts per source category')
    parser.add_argument(
        '-r', '--ranking-data', required=True,
        help='Ranking stat pdf XML file to be used for f/b')
    parser.add_argument(
        '-c', '--config', required=True,
        help='Config file for extra parameters')
    parser.add_argument(
        '-o', '--output', required=True,
        help='Output file to store mean and covariance values in JSON format')
    parser.add_argument(
        '-s', '--trigger-db', required=True,
        help='Gstlal trigger sqlite database')
    parser.add_argument(
        '-f', '--data-history', default=None,
        help='The FGMC data history as astropy table, possibly from previous runs')
    args = parser.parse_args()
    config = ConfigParser()
    config.read(args.config)
    conn = sqlite3.connect(args.trigger_db)
    activation_counts_table = Table.read(args.input, format="ascii")
    ln_likelihood_ratio_threshold = config.getfloat('p_astro',
                                                    'ln_likelihood_threshold')
    end_time_threshold = config.getfloat('p_astro', 'end_time_threshold')
    mass_threshold = config.getfloat('p_astro', 'mass_threshold')
    num_svd_bins = config.getint('p_astro', 'num_svd_bins')
    prior_type = config.get('p_astro', 'prior_type')
    background_ln_likelihood_ratios, zerolag_ln_likelihood_ratios, \
    svd_banks = _load_search_results(conn,
                                     end_time=end_time_threshold,
                                     mass=mass_threshold,
                                     ln_likelihood_ratio_threshold=
                                     ln_likelihood_ratio_threshold)

    with open(args.ranking_data, 'rb') as f:
        ranking_data_bytes = f.read()
    ln_f_over_b = _get_ln_f_over_b(ranking_data_bytes,
                                   zerolag_ln_likelihood_ratios,
                                   ln_likelihood_ratio_threshold)
    counts_instance = \
        _get_counts_instance(ln_f_over_b=ln_f_over_b,
                            svd_bank_nums=svd_banks,
                            num_svd_bins=num_svd_bins,
                            activation_counts_table=activation_counts_table,
                            lnl_threshold=ln_likelihood_ratio_threshold,
                            fgmc_data_history=args.data_history,
                            prior_type=prior_type)

    mean_values_dict = {}
    for category in ["counts_BNS",
                     "counts_NSBH",
                     "counts_BBH",
                     "counts_MassGap",
                     "counts_Terrestrial"]:

        mean_values_dict[category] = \
            counts_instance.getOneDimMean(category=category)

    with open(args.output, 'w') as f:
        content = mean_values_dict
        json.dump(content, f, default=_default)

def _default(o):
    if isinstance(o, np.int64):
        return int(o)  
    raise TypeError

def histogram_by_bin():
    parser = ArgumentParser(
        description='Executable to histogram triggers by SVD bin number')

    parser.add_argument(
        '-i', '--input',
        help='Astropy table containing  data extracted from injection database')
    parser.add_argument(
        '-o', '--output',
        help='Text file to store activation counts per bin and source category')
    parser.add_argument(
        '-c', '--config', required=True,
        help='Config file with additional parameters')

    args = parser.parse_args()

    config = ConfigParser()
    config.read(args.config)

    cfar_threshold = eval(config.get('p_astro','cfar_threshold'))
    num_svds = config.getint('p_astro','num_svd_bins')

    data = Table.read(args.input,format="ascii")
    cfar = data["cfar"]
    Gamma1 = data["Gamma1"]
    mass1 = np.maximum(data["inj_m1"], data["inj_m2"])
    mass2 = np.minimum(data["inj_m1"], data["inj_m2"])
    # definition of NS: m_2 < 3 m_sun
    # definition of BH: m_2 > 5 m_sun
    # definition of MG: m1 or m2, > 3 and <= 5 
    bns_mask = np.logical_and(mass1 <= 3.0, mass2 <= 3.0)
    nsbh_mask = np.logical_and(mass1 > 3.0, mass2 <= 3.0)
    bbh_mask = np.logical_and(mass1 > 5.0, mass2 > 5.0)
    mg_mask = np.logical_or(np.logical_and(mass1 > 3.0, mass1 <= 5.0),
                            np.logical_and(mass2 > 3.0, mass2 <= 5.0))
    cfar_mask = cfar < cfar_threshold
    # calculate activation counts
    # BNS
    svd_bank_nums_bns = Gamma1[cfar_mask * bns_mask]
    act_counts_bns = np.zeros(num_svds, dtype=int)
    _uniq, _count = np.unique(svd_bank_nums_bns, return_counts=True)
    act_counts_bns[_uniq.astype(int)] = _count
    # NSBH
    svd_bank_nums_nsbh = Gamma1[cfar_mask * nsbh_mask]
    act_counts_nsbh = np.zeros(num_svds, dtype=int)
    _uniq, _count = np.unique(svd_bank_nums_nsbh, return_counts=True)
    act_counts_nsbh[_uniq.astype(int)] = _count
    # BBH
    svd_bank_nums_bbh = Gamma1[cfar_mask * bbh_mask]
    act_counts_bbh = np.zeros(num_svds, dtype=int)
    _uniq, _count = np.unique(svd_bank_nums_bbh, return_counts=True)
    act_counts_bbh[_uniq.astype(int)] = _count
    # MG
    svd_bank_nums_mg = Gamma1[cfar_mask * mg_mask]
    act_counts_mg = np.zeros(num_svds, dtype=int)
    _uniq, _count = np.unique(svd_bank_nums_mg, return_counts=True)
    act_counts_mg[_uniq.astype(int)] = _count
    # write activation counts to astropy table
    dat = np.array([act_counts_bns, act_counts_nsbh, act_counts_bbh, act_counts_mg]).T
    t = Table(data=dat, names=("bns","nsbh","bbh","mg"))
    t.write(args.output, format="ascii")
